# Sketch - A Python-based interactive drawing program
# Copyright (C) 2003, 2004 by Bernhard Herzog
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Test cases for Sketch/Editor/doceditor.py
"""

__version__ = "$Revision: 1.13 $"
# $Source: /cvs/gnome/sketch/test/test_doceditor.py,v $
# $Id: test_doceditor.py,v 1.13 2004/12/29 22:26:08 bherzog Exp $

import sys
import unittest
import traceback

import support
support.add_sketch_dir_to_path()

# init_base is necessary so that we can use text objects since they need
# fonts.
support.init_base()

import Sketch
from Sketch import Point, Rect, Trafo, CreatePath, ContAngle
from Sketch.Base.const import HANDLES, CURRENTINFO, OBJECT_CHANGED, \
     CHILDREN, INSERTED, REMOVED, REDRAW
from Sketch.Graphics import PolyBezier, Layer, Rectangle, Group
from Sketch.Editor import EditorWithSelection, InteractiveDocument, Context, \
     toolmap, Button1Mask, ShiftMask
from Sketch.Editor.tools import ToolInstance, ToolInfo


class TestEditorWithSelection(unittest.TestCase):

    """Simple tests for EditorWithSelection"""

    def setUp(self):
        """Override to create a document and editor

        Bind the document to self.document and the editor to self.editor
        """
        self.document = InteractiveDocument(create_layer = 1)
        self.editor = EditorWithSelection(self.document)

    def tearDown(self):
        """Destroy the editor object"""
        self.editor.Destroy()

    def test_initial_settings(self):
        """Test EditorWithSelection initial settings"""
        editor = self.editor

        self.assert_(editor.Document() is self.document)

        # No snapping by default
        self.failIf(editor.IsSnappingToGrid())
        self.failIf(editor.IsSnappingToGuides())
        self.failIf(editor.IsSnappingToObjects())

        # No selection yet
        self.failIf(editor.HasSelection())

        # There's no selection so there shouldn't be any handles either
        self.assertEquals(editor.Handles(), [])



class DocEditorTestsWithMessages(unittest.TestCase):

    """Base class for EditorWithSelection tests that involve messages"""

    subscribed_channels = ()

    def setUp(self):
        """Create a document and an editor and subscribe some messages

        The document is an instance of InteractiveDocument and is bound
        to self.document.  The document contains a layer which is
        available as self.layer.  The editor is bound to self.editor and
        is an instance of EditorWithSelection.

        After these objects have been created setUp subscrubes
        self.receiver to messages as described by the
        subscribed_channels class variable.  This variable is a sequence
        of (attr, channels) pairs with attr being the name of an
        instance attribute and channels a sequence of channel names.
        For all the channels self.receiver is subscribed once to the
        attribute of self called attr (so that an attr value 'document'
        refers to self.document).  The channel name will be passed to
        self.receiver as well as the last positional argument.
        self.receiver records all messages in self.received_messages
        (see receiver's doc string for more details).  The test methods
        can check that variable to verify which messages have been sent
        since the call to setUp.
        """
        self.document = InteractiveDocument(create_layer = 1)
        self.layer = self.document.AppendLayer("test")
        self.editor = EditorWithSelection(self.document)
        for attr, channels in self.subscribed_channels:
            for channel in channels:
                getattr(self, attr).Subscribe(channel, self.receiver, channel)
        self.received_messages = []

    def tearDown(self):
        """Clean up resources initialized in setup.

        The cleanup destros the editor object and removes the references
        to the document, the editor and the layer, and clears the
        received_messages list.
        """
        self.received_messages = []
        if self.editor is not None:
            self.editor.Destroy()
        self.editor = self.document = self.layer = None

    def receiver(self, *args):
        """Generic receiver for connector messages.

        Simply append the argument tuple to self.received_messages().
        """
        self.received_messages.append(args)


class MockToolInstance(ToolInstance):

    """
    ToolInstance that records whether some of the methods have been called
    """

    def __init__(self, *args):
        ToolInstance.__init__(self, *args)
        self.called_methods = []

    def DelayButtonPress(self):
        self.called_methods.append(("DelayButtonPress",))
        return 1

    def ButtonPress(self, canvas, p, snapped, button, state, handle = None):
        ToolInstance.ButtonPress(self, canvas, p, snapped, button, state,
                                 handle = None)
        self.called_methods.append(("ButtonPress", canvas, p, snapped, button,
                                    state, handle))

    def PointerMotion(self, canvas, p, snapped, state):
        ToolInstance.PointerMotion(self, canvas, p, snapped, state)
        self.called_methods.append(("PointerMotion", canvas, p, snapped,
                                    state))

    def ButtonRelease(self, canvas, p, snapped, button, state):
        ToolInstance.ButtonRelease(self, canvas, p, snapped, button, state)
        self.called_methods.append(("ButtonRelease", canvas, p, snapped,
                                    button, state))

    def ButtonClick(self, canvas, p, snapped, button, state, handle = None):
        ToolInstance.ButtonClick(self, canvas, p, snapped, button, state,
                                 handle = None)
        self.called_methods.append(("ButtonClick", canvas, p, snapped,
                                    button, state, handle))

    def End(self):
        """Extend the inherited method to record that it has been called."""
        ToolInstance.End(self)
        self.called_methods.append(("End",))

MockTool = ToolInfo("MockTool", MockToolInstance, cursor = 'CurZoom')
MockTempTool = ToolInfo("MockTempTool", MockToolInstance, cursor = 'CurPick')

class MockCanvas:

    """Mock canvas the does nothing"""

    def max_snap_distance(self):
        return 1

    def __dummy(self, *args, **kw):
        pass

    hide_handles = __dummy
    show_handles = __dummy

    def inverting_device(self):
        return MockGC()

    def test_device(self):
        # The MockGC isn't reall right but the only test that needs it
        # currentyl is the very simple selection tool test which should
        # get by with this.
        return MockGC()


class MockGC:

    """Very simple Mock GC

    Objects of this class allow arbitrary methods to be called with
    arbitrary arguments.
    """

    def __getattr__(self, methodname):
        def dummy(*args, **kw):
            pass
        return dummy

    def HitRectAroundPoint(self, p):
        rad = 1
        return Rect(p.x - rad, p.y - rad, p.x + rad, p.y + rad)



class TestEditorWithSelectionTools(DocEditorTestsWithMessages):

    """EditorWithSelection test that involve the MockTool"""

    subscribed_channels = [("editor", [HANDLES, CURRENTINFO])]

    def setUp(self):
        """Prepare for a unit test

        Create a document as self.document and an editor as self.editor.
        Subscribe to some editor channels.

        Store a copy of the toolmap and add the MockTool to the toolmap.
        Initialize the received_messages list.
        """
        DocEditorTestsWithMessages.setUp(self)
        self.old_toolmap = toolmap.copy()
        toolmap["MockTool"] = MockTool

    def tearDown(self):
        """Restore the toolmap and destroy the editor object"""
        toolmap.clear()
        toolmap.update(self.old_toolmap)
        DocEditorTestsWithMessages.tearDown(self)

    def test_set_tool(self):
        """Test EditorWithSelection.SetTool"""
        editor = self.editor

        # Select the mock tool, get a reference to the actual instance
        # used and select another tool
        editor.SetTool("MockTool")
        mock_tool = editor.Tool()
        self.assert_(isinstance(mock_tool, MockToolInstance))
        editor.SetTool("SelectionTool")

        # Now the mock tool's End method should have been called and a
        # HANDLES should have been emitted twice because we've changed
        # tools twice
        self.assertEquals(mock_tool.called_methods, [("End",)])
        self.assertEquals(self.received_messages, [(HANDLES,), (HANDLES,)])

    def test_mouse_interaction_drag(self):
        """Test EditorWithSelection mouse interaction: drag"""
        editor = self.editor

        # Select the mock tool, get a reference to the actual instance
        # used so that we can look at which of its methods were called
        # later.
        editor.SetTool("MockTool")
        mock_tool = editor.Tool()

        # simulate a click-drag-release
        canvas = MockCanvas()

        # Clear the messages list before each method call so that we can
        # check more precisely which method call leads to which
        # messages.
        self.received_messages = []
        editor.ButtonPress(canvas, Point(10, 20), 1, 0)
        self.assertEquals(self.received_messages, [(CURRENTINFO,)])

        self.received_messages = []
        editor.PointerMotion(canvas, Point(20, 20), Button1Mask)
        self.assertEquals(self.received_messages, [(CURRENTINFO,)])

        self.received_messages = []
        editor.ButtonRelease(canvas, Point(25, 28), 1, Button1Mask)
        # The ButtonRelease call should result in at least CURRENTINFO
        # and perhaps HANDLES but that doesn't happen at the moment
        # which is a bug.
        self.assertEquals(self.received_messages, [])
        #[(CURRENTINFO,), (HANDLES,)])

        self.assertEquals(mock_tool.called_methods,
                          [("ButtonPress",
                            canvas, Point(10, 20), Point(10, 20), 1, 0, None),
                           ("PointerMotion",
                            canvas, Point(20, 20), Point(20, 20), Button1Mask),
                           ("ButtonRelease",
                            canvas, Point(25,28), Point(25,28), 1,Button1Mask),
                          ])

    def test_mouse_interaction_click(self):
        """Test EditorWithSelection mouse interaction: click"""
        editor = self.editor

        # Select the mock tool, get a reference to the actual instance
        # used so that we can look at which of its methods were called
        # later.
        editor.SetTool("MockTool")
        mock_tool = editor.Tool()

        # simulate a click-drag-release
        canvas = MockCanvas()

        # Clear the messages list before each method call so that we can
        # check more precisely which method call leads to which
        # messages.
        self.received_messages = []
        editor.ButtonClick(canvas, Point(10, 20), 1, ShiftMask)
        self.assertEquals(self.received_messages, [])

        self.assertEquals(mock_tool.called_methods,
                          [("ButtonClick",
                            canvas, Point(10, 20), Point(10, 20), 1, ShiftMask,
                            None),
                          ])

    def test_set_tool_with_unfinished_poly_line(self):
        """Test EditorWithSelection.SetTool() with unfinished poly line

        Test for savannah bugs #3760 and #3918.  applies to both
        PolyBezierToolInstance and PolyLineToolInstance.
        """
        editor = self.editor
        canvas = MockCanvas()

        # Start creating a poly line with a singe click-drag-release
        editor.SetTool("CreatePoly")
        editor.ButtonPress(canvas, Point(0, 0), 1, 0)
        editor.PointerMotion(canvas, Point(10, 10), Button1Mask)
        editor.ButtonRelease(canvas, Point(10, 10), 1, Button1Mask)

        # Now we have an unfinished polyline. It finished enough to be
        # inserted into the document but the editor still accepts
        # additonal mouse interaction.
        #
        # In this state we switch the tools.  This should result in the
        # poly line being inserted into the document.  The bug was an
        # exception during this.
        editor.SetTool("CreateRect")

        # If all went well the object has been inserted and is now the
        # selected object
        obj = self.editor.CurrentObject()
        self.failIf(obj is None)
        self.failUnless(obj.is_Bezier)

    def test_create_poly_tool_single(self):
        """Test EditorWithSelection and CreatePoly single press/drag/release
        """
        editor = self.editor
        canvas = MockCanvas()

        # Start creating a poly line with a single click-drag-release
        editor.SetTool("CreatePoly")
        editor.ButtonPress(canvas, Point(0, 0), 1, 0)
        editor.PointerMotion(canvas, Point(10, 10), Button1Mask)
        editor.ButtonRelease(canvas, Point(10, 10), 1, Button1Mask)

        # Middle button click finishes drawing
        editor.ButtonClick(canvas, Point(100, 100), 2, 0)

        # If all went well the object has been inserted and is now the
        # selected object
        obj = self.editor.CurrentObject()
        self.failIf(obj is None)
        self.failUnless(obj.is_Bezier)
        self.assertEquals(len(obj.Paths()), 1)
        self.assertEquals(obj.Paths()[0].get_save(),
                          [(0, 0, ContAngle),
                           (10, 10, ContAngle)])

    def test_create_poly_tool_click(self):
        """Test EditorWithSelection and CreatePoly starting with click
        """
        editor = self.editor
        canvas = MockCanvas()

        # Start creating a poly line with a click for the first point
        # and a click-drag-release for the second.  Then finish with a
        # middle button click
        editor.SetTool("CreatePoly")
        editor.ButtonPress(canvas, Point(0, 0), 1, 0)
        editor.ButtonRelease(canvas, Point(0, 0), 1, Button1Mask)
        editor.ButtonPress(canvas, Point(8, 2), 1, 0)
        editor.PointerMotion(canvas, Point(9, 1), Button1Mask)
        editor.ButtonRelease(canvas, Point(10, 0), 1, Button1Mask)
        editor.ButtonClick(canvas, Point(100, 100), 2, 0)

        # If all went well the object has been inserted and is now the
        # selected object
        obj = self.editor.CurrentObject()
        self.failIf(obj is None)
        self.failUnless(obj.is_Bezier)
        self.assertEquals(len(obj.Paths()), 1)
        self.assertEquals(obj.Paths()[0].get_save(),
                          [(0, 0, ContAngle),
                           (10, 0, ContAngle)])

    def test_create_poly_tool_multiple(self):
        """Test EditorWithSelection and CreatePoly multiple press/drag/release
        """
        editor = self.editor
        canvas = MockCanvas()

        # Start creating a poly line with a singe click-drag-release
        editor.SetTool("CreatePoly")
        editor.ButtonPress(canvas, Point(0, 0), 1, 0)
        editor.PointerMotion(canvas, Point(10, 0), Button1Mask)
        editor.ButtonRelease(canvas, Point(10, 0), 1, Button1Mask)
        editor.ButtonPress(canvas, Point(20, 20), 1, 0)
        editor.PointerMotion(canvas, Point(20, 10), Button1Mask)
        editor.ButtonRelease(canvas, Point(10, 20), 1, Button1Mask)
        editor.ButtonClick(canvas, Point(100, 100), 2, 0)

        # If all went well the object has been inserted and is now the
        # selected object
        obj = self.editor.CurrentObject()
        self.failIf(obj is None)
        self.failUnless(obj.is_Bezier)
        self.assertEquals(len(obj.Paths()), 1)
        self.assertEquals(obj.Paths()[0].get_save(),
                          [(0, 0, ContAngle),
                           (10, 0, ContAngle),
                           (10, 20, ContAngle)])


    def test_edit_simple(self):
        """Test EditorWithSelection with EditTool (very simple)"""
        editor = self.editor
        canvas = MockCanvas()

        path = CreatePath()
        path.AppendLine(0, 0)
        path.AppendLine(10, 10)
        poly = PolyBezier(paths = (path,))

        self.document.Insert(poly, self.layer)
        self.editor.SelectObject(poly)

        editor.SetTool("EditTool")

        # Get a handle so that we can drag it
        handle = editor.Handles()[0]

        # Now drag it to (5, 0)
        editor.ButtonPress(canvas, Point(0, 0), 1, 0, handle=handle)
        editor.PointerMotion(canvas, Point(2, 0), Button1Mask)
        editor.ButtonRelease(canvas, Point(5, 0), 1, Button1Mask)

        # Check the new state of the poly line
        obj = self.editor.CurrentObject()
        self.failIf(obj is None)
        self.failUnless(obj.is_Bezier)
        self.assertEquals(len(obj.Paths()), 1)
        self.assertEquals(obj.Paths()[0].get_save(),
                          [(5, 0, ContAngle),
                           (10, 10, ContAngle)])

    def test_select_simple(self):
        """Test EditorWithSelection with SelectTool (very simple)"""
        editor = self.editor
        canvas = MockCanvas()

        path = CreatePath()
        path.AppendLine(0, 0)
        path.AppendLine(10, 10)
        poly = PolyBezier(paths = (path,))

        self.document.Insert(poly, self.layer)

        editor.SetTool("SelectionTool")

        # Select with rubber banding.  Since the default join is mitred
        # join, the bounding rect of poly is quite a bit larger (by more
        # than 5pt) than (0, 0, 10, 10), so make sure that start and
        # stop point define a sufficiently large rectangle.
        editor.ButtonPress(canvas, Point(-10, -10), 1, 0)
        editor.PointerMotion(canvas, Point(5, 0), Button1Mask)
        editor.ButtonRelease(canvas, Point(20, 25), 1, Button1Mask)

        # Check the new state of the poly line
        obj = self.editor.CurrentObject()
        self.failUnless(obj is poly)


    def test_text_simple(self):
        """Test EditorWithSelection with TextTool (very simple)"""
        editor = self.editor
        canvas = MockCanvas()

        editor.SetTool("TextTool")

        # Click somewhere on the document to start creating a new text
        # object
        editor.ButtonPress(canvas, Point(100, 100), 1, 0)
        editor.ButtonRelease(canvas, Point(100, 100), 1, Button1Mask)

        # Now type some text
        tool = editor.Tool()
        for c in "Skencil":
            tool.CallObjectEditorMethod(Sketch.Editor.CommonTextEditor,
                                        'InsertCharacter', c)

        # Now select a different tool, e.g. the selection tool.  Text
        # editing should be finished and the new text object the
        # selected object
        editor.SetTool("SelectionTool")

        obj = editor.CurrentObject()
        self.assert_(obj.is_Text)
        self.assertEquals(obj.Text(), "Skencil")


class TestEditorObjectRearrangement(DocEditorTestsWithMessages):

    """EditorWithSelection tests that rearrange and duplicate objects"""

    def test_move_up_simple(self):
        """Test EditorWithSelection.MoveSelectedUp() very simple

        Test a very simple version of MoveSelectedUp: two objects on one
        layer.  The bottom one is selected and then moved up.  Test
        whether the order of the objects is correct after that and
        whether undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)

        self.editor.SelectObject(r1)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # The actual operation to test
        self.editor.MoveSelectedUp()

        # Now r1 should be above r2
        self.assertEquals(self.layer.GetObjects(), [r2, r1])

        # r1 should still be selected
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the order r1 and r2 and the selection of the editor should
        # still be r1
        self.assertEquals(self.layer.GetObjects(), [r1, r2])
        self.assertEquals(self.editor.SelectedObjects(), [r1])

    def test_move_up_several(self):
        """Test EditorWithSelection.MoveSelectedUp() several objects at once

        Several objects on one layer.  Two immediately above each other
        but below the topmost object are selected and then moved up.
        This was chosen so that they actually would move up and to test
        that their relative order doesn't change.  Test whether the
        order of the objects is correct after that and whether undo
        works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(r3, self.layer)

        self.editor.SelectObject([r1, r2])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # The actual operation to test
        self.editor.MoveSelectedUp()

        # Now r1 and r2 should be above r3 and r2 still above r1
        self.assertEquals(self.layer.GetObjects(), [r3, r1, r2])

        # The selection must not have changed
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the order r1, r2, r3 and, again, the selection must not have
        # changed
        self.assertEquals(self.layer.GetObjects(), [r1, r2, r3])
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

    def test_move_up_all_children(self):
        """Test EditorWithSelection.MoveSelectedUp() all objects in layer

        Two objects on one layer.  Both are selected and then moved up.
        This is a special case in which no reordering must take place
        because the topmost object can't be moved up and if the next
        lower object would were moved up the relative order of the
        selected objects would change.  Test that indeed nothing
        happens.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)

        self.editor.SelectObject([r1, r2])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # The actual operation to test
        self.editor.MoveSelectedUp()

        # Since all objects in the layer are selected, nothing should
        # have been done
        self.assertEquals(self.layer.GetObjects(), [r1, r2])

        # The selection should still be the same
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

    def test_move_up_in_several_layers(self):
        """Test EditorWithSelection.DuplicateSelected() with multiple layers

        Two different layers with several objects each.  On each layer
        several objects are seleced and then moved up.  Test the order
        of the objects afterwards and whether undo works.
        """
        r11 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r12 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r13 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(r11, self.layer)
        self.document.Insert(r12, self.layer)
        self.document.Insert(r13, self.layer)

        new_layer = self.document.AppendLayer("new layer")
        r21 = Rectangle(Trafo(10, 0, 0, 10, 100, 0))
        r22 = Rectangle(Trafo(10, 0, 0, 10, 100, 100))
        r23 = Rectangle(Trafo(10, 0, 0, 10, 100, 200))
        self.document.Insert(r21, new_layer)
        self.document.Insert(r22, new_layer)
        self.document.Insert(r23, new_layer)

        self.editor.SelectObject([r11, r13, r21, r22])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r21, r22])

        # The actual operation to test
        self.editor.MoveSelectedUp()

        # Check the order of the children of self.layer and new_layer
        self.assertEquals(self.layer.GetObjects(), [r12, r11, r13])
        self.assertEquals(new_layer.GetObjects(), [r23, r21, r22])
        # the selection should not have changed.
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r21, r22])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layers should be in the
        # same order as before and the selection should still be the
        # same
        self.assertEquals(self.layer.GetObjects(), [r11, r12, r13])
        self.assertEquals(new_layer.GetObjects(), [r21, r22, r23])
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r21, r22])


    def test_move_up_in_group(self):
        """Test EditorWithSelection.MoveSelectedUp() objects in a group

        A group with two objects.  The lower one is selected and moved
        up.  Check the order of the children afterwards and whether undo
        works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        group = Group([r1, r2])
        self.document.Insert(group, self.layer)

        self.editor.SelectObject(r1)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # The actual operation to test
        self.editor.MoveSelectedUp()

        # The order of the children of the group should have changed
        # now.
        self.assertEquals(group.GetObjects(), [r2, r1])

        # The selection should still be the same, though.
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the order of children of the group is once
        # again r1, r2 and the selection of the editor is still the same
        # as before.
        self.assertEquals(group.GetObjects(), [r1, r2])
        self.assertEquals(self.editor.SelectedObjects(), [r1])

    def test_move_down_simple(self):
        """Test EditorWithSelection.MoveSelectedDown() very simple

        Test a very simple version of MoveSelectedDown: two objects on
        one layer.  The top one is selected and then moved down.  Test
        whether the order of the objects is correct after that and
        whether undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)

        self.editor.SelectObject(r2)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r2])

        # The actual operation to test
        self.editor.MoveSelectedDown()

        # Now r2 should be below r1
        self.assertEquals(self.layer.GetObjects(), [r2, r1])

        # r2 should still be selected
        self.assertEquals(self.editor.SelectedObjects(), [r2])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the order r1 and r2 and the selection of the editor should
        # still be r2
        self.assertEquals(self.layer.GetObjects(), [r1, r2])
        self.assertEquals(self.editor.SelectedObjects(), [r2])

    def test_move_down_several(self):
        """Test EditorWithSelection.MoveSelectedDown() several objects at once

        Several objects on one layer.  Two immediately above each other
        but above the lowermost object are selected and then moved down.
        This was chosen so that they actually would move down and to
        test that their relative order doesn't change.  Test whether the
        order of the objects is correct after that and whether undo
        works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(r3, self.layer)

        self.editor.SelectObject([r2, r3])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r2, r3])

        # The actual operation to test
        self.editor.MoveSelectedDown()

        # Now r1 should be above r2 and r3 and r3 still above r2
        self.assertEquals(self.layer.GetObjects(), [r2, r3, r1])

        # The selection should not have changed
        self.assertEquals(self.editor.SelectedObjects(), [r2, r3])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the order r1, r2, r3 and the selection should still not have
        # changed.
        self.assertEquals(self.layer.GetObjects(), [r1, r2, r3])
        self.assertEquals(self.editor.SelectedObjects(), [r2, r3])

    def test_move_down_all_children(self):
        """Test EditorWithSelection.MoveSelectedDown() all objects in layer

        Two objects on one layer.  Both are selected and then moved
        down.  This is a special case in which no reordering must take
        place because the lowermost object can't be moved down and if
        the next higher object were moved down the relative order of the
        selected objects would change.  Test that indeed nothing
        happens.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)

        self.editor.SelectObject([r1, r2])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # The actual operation to test
        self.editor.MoveSelectedDown()

        # Since all objects in the layer are selected, nothing should
        # have been done
        self.assertEquals(self.layer.GetObjects(), [r1, r2])

        # The selection should still be the same
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

    def test_move_down_in_several_layers(self):
        """Test EditorWithSelection.MoveSelectedDown() with multiple layers

        Two different layers with several objects each.  On each layer
        several objects are seleced and then moved down.  Test the order
        of the objects afterwards and whether undo works.
        """
        r11 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r12 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r13 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(r11, self.layer)
        self.document.Insert(r12, self.layer)
        self.document.Insert(r13, self.layer)

        new_layer = self.document.AppendLayer("new layer")
        r21 = Rectangle(Trafo(10, 0, 0, 10, 100, 0))
        r22 = Rectangle(Trafo(10, 0, 0, 10, 100, 100))
        r23 = Rectangle(Trafo(10, 0, 0, 10, 100, 200))
        self.document.Insert(r21, new_layer)
        self.document.Insert(r22, new_layer)
        self.document.Insert(r23, new_layer)

        self.editor.SelectObject([r11, r13, r22, r23])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r22, r23])

        # The actual operation to test
        self.editor.MoveSelectedDown()

        # Check the order of the children of self.layer and new_layer
        self.assertEquals(self.layer.GetObjects(), [r11, r13, r12])
        self.assertEquals(new_layer.GetObjects(), [r22, r23, r21])
        # the selection should not have changed.
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r22, r23])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layers should be in the
        # same order as before and the selection should still be the
        # same
        self.assertEquals(self.layer.GetObjects(), [r11, r12, r13])
        self.assertEquals(new_layer.GetObjects(), [r21, r22, r23])
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r22, r23])

    def test_move_down_in_group(self):
        """Test EditorWithSelection.MoveSelectedDown() objects in a group

        A group with two objects.  The upper one is selected and moved
        down.  Check the order of the children afterwards and whether
        undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        group = Group([r1, r2])
        self.document.Insert(group, self.layer)

        self.editor.SelectObject(r2)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r2])

        # The actual operation to test
        self.editor.MoveSelectedDown()

        # The order of the children of the group should have changed
        # now.
        self.assertEquals(group.GetObjects(), [r2, r1])

        # The selection should still be the same, though.
        self.assertEquals(self.editor.SelectedObjects(), [r2])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the order of children of the group is once
        # again r1, r2 and the selection of the editor is still the same
        # as before.
        self.assertEquals(group.GetObjects(), [r1, r2])
        self.assertEquals(self.editor.SelectedObjects(), [r2])

    def test_move_to_top_simple(self):
        """Test EditorWithSelection.MoveSelectedToTop() very simple

        Three objects on one layer.  The bottom one is selected and then
        moved up.  Test whether the object is at the top afterwards and
        whether undo works.
        """
        # Note: Make sure that this testcase can't be fulfilled with the
        # semantics of MoveUp so we have three objects and select he
        # lowermost one.
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(r3, self.layer)

        self.editor.SelectObject(r1)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # The actual operation to test
        self.editor.MoveSelectedToTop()

        # r1 should be on top now
        self.assertEquals(self.layer.GetObjects(), [r2, r3, r1])

        # The selection should not have changed
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the order r1, r2, r3 and the selection of the editor should
        # still be the same as before
        self.assertEquals(self.layer.GetObjects(), [r1, r2, r3])
        self.assertEquals(self.editor.SelectedObjects(), [r1])

    def test_move_to_top_several(self):
        """Test EditorWithSelection.MoveSelectedToTop() several objects

        Several objects on one layer.  Two objects not including the
        topmost one are selected and then moved to the top.  This was
        chosen so that they actually would move up and to test that
        their relative order doesn't change.  Test the order of the
        objects is correct afterwards and whether undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        r4 = Rectangle(Trafo(10, 0, 0, 10, 0, 400))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(r3, self.layer)
        self.document.Insert(r4, self.layer)

        self.editor.SelectObject([r1, r2])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # The actual operation to test
        self.editor.MoveSelectedToTop()

        # r1 and r2 should be at the top.  The relative order of the
        # other objects should not have changed.
        self.assertEquals(self.layer.GetObjects(), [r3, r4, r1, r2])

        # The selection should not have changed
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the original order and the selection should still not have
        # changed.
        self.assertEquals(self.layer.GetObjects(), [r1, r2, r3, r4])
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

    def test_move_to_top_all_children(self):
        """Test EditorWithSelection.MoveSelectedToTop() all objects in layer

        Two objects on one layer.  Both are selected and then moved to
        the top.  This is a special case in which no reordering must
        take place because the none of the objects can be moved without
        changing their relative stacking order.  Test that indeed
        nothing happens.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)

        self.editor.SelectObject([r1, r2])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # The actual operation to test
        self.editor.MoveSelectedToTop()

        # Since all objects in the layer are selected, nothing should
        # have been done
        self.assertEquals(self.layer.GetObjects(), [r1, r2])

        # The selection should still be the same
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

    def test_move_to_top_in_several_layers(self):
        """Test EditorWithSelection.MoveSelectedToTop() with multiple layers

        Two different layers with several objects each.  On each layer
        several objects are seleced and then moved to the top.  Test the
        order of the objects afterwards and whether undo works.
        """
        r11 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r12 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r13 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        r14 = Rectangle(Trafo(10, 0, 0, 10, 0, 300))
        self.document.Insert(r11, self.layer)
        self.document.Insert(r12, self.layer)
        self.document.Insert(r13, self.layer)
        self.document.Insert(r14, self.layer)

        new_layer = self.document.AppendLayer("new layer")
        r21 = Rectangle(Trafo(10, 0, 0, 10, 100, 0))
        r22 = Rectangle(Trafo(10, 0, 0, 10, 100, 100))
        r23 = Rectangle(Trafo(10, 0, 0, 10, 100, 200))
        r24 = Rectangle(Trafo(10, 0, 0, 10, 100, 300))
        self.document.Insert(r21, new_layer)
        self.document.Insert(r22, new_layer)
        self.document.Insert(r23, new_layer)
        self.document.Insert(r24, new_layer)

        self.editor.SelectObject([r11, r13, r22, r23])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r22, r23])

        # The actual operation to test
        self.editor.MoveSelectedToTop()

        # Check the order of the children of self.layer and new_layer
        self.assertEquals(self.layer.GetObjects(), [r12, r14, r11, r13])
        self.assertEquals(new_layer.GetObjects(), [r21, r24, r22, r23])
        # the selection should not have changed.
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r22, r23])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layers should be in the
        # same order as before and the selection should still be the
        # same
        self.assertEquals(self.layer.GetObjects(), [r11, r12, r13, r14])
        self.assertEquals(new_layer.GetObjects(), [r21, r22, r23, r24])
        self.assertEquals(self.editor.SelectedObjects(), [r11, r13, r22, r23])

    def test_move_to_top_in_group(self):
        """Test EditorWithSelection.MoveSelectedToTop() objects in a group

        A group with two objects.  The upper one is selected and moved
        down.  Check the order of the children afterwards and whether
        undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        group = Group([r1, r2, r3])
        self.document.Insert(group, self.layer)

        self.editor.SelectObject(r1)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # The actual operation to test
        self.editor.MoveSelectedToTop()

        # The order of the children of the group should have changed
        # now.
        self.assertEquals(group.GetObjects(), [r2, r3, r1])

        # The selection should still be the same, though.
        self.assertEquals(self.editor.SelectedObjects(), [r1])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the group are again in their
        # original order and the selection of the editor is still the
        # same as before.
        self.assertEquals(group.GetObjects(), [r1, r2, r3])
        self.assertEquals(self.editor.SelectedObjects(), [r1])

    def test_move_to_top_mixed(self):
        """Test EditorWithSelection.MoveSelectedToTop() in groups and layers

        A trickier case: A layer with 4 rectanges r1, ..., r4 and a
        group G with three rectangles [r21, r22, r23].  The children of
        the layer are [r1, r2, G, r3, r4].  The selection is [r1, r21,
        r3].  The selection is moved to the top.  Afterwards, The order
        in the layer must be [r2, G, r4, r1, r3] and in the group [r22,
        r23, r21].
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        r4 = Rectangle(Trafo(10, 0, 0, 10, 0, 300))
        r21 = Rectangle(Trafo(10, 0, 0, 10, 100, 0))
        r22 = Rectangle(Trafo(10, 0, 0, 10, 100, 100))
        r23 = Rectangle(Trafo(10, 0, 0, 10, 100, 200))
        group = Group([r21, r22, r23])
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(group, self.layer)
        self.document.Insert(r3, self.layer)
        self.document.Insert(r4, self.layer)

        self.editor.SelectObject([r1, r21, r3])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r21, r3])

        # The actual operation to test
        self.editor.MoveSelectedToTop()

        # The order of the children of the group should have changed
        # now.
        self.assertEquals(group.GetObjects(), [r22, r23, r21])
        self.assertEquals(self.layer.GetObjects(), [r2, group, r4, r1, r3])

        # The selection should still be the same, though.
        self.assertEquals(self.editor.SelectedObjects(), [r1, r21, r3])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer and the group are
        # again in their original order and the selection of the editor
        # is still the same as before.
        self.assertEquals(group.GetObjects(), [r21, r22, r23])
        self.assertEquals(self.layer.GetObjects(), [r1, r2, group, r3, r4])
        self.assertEquals(self.editor.SelectedObjects(), [r1, r21, r3])

    def test_move_to_bottom_simple(self):
        """Test EditorWithSelection.MoveSelectedToBottom() very simple

        Three objects on one layer.  The top one is selected and then
        moved up.  Test whether the object is at the bottom afterwards
        and whether undo works.
        """
        # Note: Make sure that this testcase can't be fulfilled with the
        # semantics of MoveDown so we have three objects and select he
        # topmost one.
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(r3, self.layer)

        self.editor.SelectObject(r3)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r3])

        # The actual operation to test
        self.editor.MoveSelectedToBottom()

        # r3 should be on bottom now
        self.assertEquals(self.layer.GetObjects(), [r3, r1, r2])

        # The selection should not have changed
        self.assertEquals(self.editor.SelectedObjects(), [r3])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the original order and the selection of the editor should
        # still be the same as before
        self.assertEquals(self.layer.GetObjects(), [r1, r2, r3])
        self.assertEquals(self.editor.SelectedObjects(), [r3])

    def test_move_to_bottom_several(self):
        """Test EditorWithSelection.MoveSelectedToBottom() several objects

        Several objects on one layer.  Two objects not including the
        bottommost one are selected and then moved to the bottom.  This
        was chosen so that they actually would move down and to test
        that their relative order doesn't change.  Test the order of the
        objects is correct afterwards and whether undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        r4 = Rectangle(Trafo(10, 0, 0, 10, 0, 400))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(r3, self.layer)
        self.document.Insert(r4, self.layer)

        self.editor.SelectObject([r3, r4])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r3, r4])

        # The actual operation to test
        self.editor.MoveSelectedToBottom()

        # r3 and r4 should be at the bottom.  The relative order of the
        # other objects should not have changed.
        self.assertEquals(self.layer.GetObjects(), [r3, r4, r1, r2])

        # The selection should not have changed
        self.assertEquals(self.editor.SelectedObjects(), [r3, r4])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer are once again in
        # the original order and the selection should still not have
        # changed.
        self.assertEquals(self.layer.GetObjects(), [r1, r2, r3, r4])
        self.assertEquals(self.editor.SelectedObjects(), [r3, r4])

    def test_move_to_bottom_all_children(self):
        """Test EditorWithSelection.MoveSelectedToBottom() all objects in layer

        Two objects on one layer.  Both are selected and then moved to
        the bottom.  This is a special case in which no reordering must
        take place because the none of the objects can be moved without
        changing their relative stacking order.  Test that indeed
        nothing happens.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)

        self.editor.SelectObject([r1, r2])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

        # The actual operation to test
        self.editor.MoveSelectedToBottom()

        # Since all objects in the layer are selected, nothing should
        # have been done
        self.assertEquals(self.layer.GetObjects(), [r1, r2])

        # The selection should still be the same
        self.assertEquals(self.editor.SelectedObjects(), [r1, r2])

    def test_move_to_bottom_in_several_layers(self):
        """Test EditorWithSelection.MoveSelectedToBottom() with multiple layers

        Two different layers with several objects each.  On each layer
        several objects are seleced and then moved to the bottom.  Test the
        order of the objects afterwards and whether undo works.
        """
        r11 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r12 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r13 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        r14 = Rectangle(Trafo(10, 0, 0, 10, 0, 300))
        self.document.Insert(r11, self.layer)
        self.document.Insert(r12, self.layer)
        self.document.Insert(r13, self.layer)
        self.document.Insert(r14, self.layer)

        new_layer = self.document.AppendLayer("new layer")
        r21 = Rectangle(Trafo(10, 0, 0, 10, 100, 0))
        r22 = Rectangle(Trafo(10, 0, 0, 10, 100, 100))
        r23 = Rectangle(Trafo(10, 0, 0, 10, 100, 200))
        r24 = Rectangle(Trafo(10, 0, 0, 10, 100, 300))
        self.document.Insert(r21, new_layer)
        self.document.Insert(r22, new_layer)
        self.document.Insert(r23, new_layer)
        self.document.Insert(r24, new_layer)

        self.editor.SelectObject([r12, r14, r22, r23])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r12, r14, r22, r23])

        # The actual operation to test
        self.editor.MoveSelectedToBottom()

        # Check the order of the children of self.layer and new_layer
        self.assertEquals(self.layer.GetObjects(), [r12, r14, r11, r13])
        self.assertEquals(new_layer.GetObjects(), [r22, r23, r21, r24])
        # the selection should not have changed.
        self.assertEquals(self.editor.SelectedObjects(), [r12, r14, r22, r23])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layers should be in the
        # same order as before and the selection should still be the
        # same
        self.assertEquals(self.layer.GetObjects(), [r11, r12, r13, r14])
        self.assertEquals(new_layer.GetObjects(), [r21, r22, r23, r24])
        self.assertEquals(self.editor.SelectedObjects(), [r12, r14, r22, r23])

    def test_move_to_bottom_in_group(self):
        """Test EditorWithSelection.MoveSelectedToBottom() objects in a group

        A group with two objects.  The upper one is selected and moved
        down.  Check the order of the children afterwards and whether
        undo works.
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        group = Group([r1, r2, r3])
        self.document.Insert(group, self.layer)

        self.editor.SelectObject(r3)

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(), [r3])

        # The actual operation to test
        self.editor.MoveSelectedToBottom()

        # The order of the children of the group should have changed
        # now.
        self.assertEquals(group.GetObjects(), [r3, r1, r2])

        # The selection should still be the same, though.
        self.assertEquals(self.editor.SelectedObjects(), [r3])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the group are again in their
        # original order and the selection of the editor is still the
        # same as before.
        self.assertEquals(group.GetObjects(), [r1, r2, r3])
        self.assertEquals(self.editor.SelectedObjects(), [r3])

    def test_move_to_bottom_mixed(self):
        """Test EditorWithSelection.MoveSelectedToBottom() in groups and layers

        A trickier case: A layer with 4 rectanges r1, ..., r4 and a
        group G with three rectangles [r21, r22, r23].  The children of
        the layer are [r1, r2, G, r3, r4].  The selection is [r2, r23,
        r4].  The selection is moved to the bottom.  Afterwards, The order
        in the layer must be [r2, r4, r1, G, r3] and in the group [r23,
        r21, r22].
        """
        r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        r4 = Rectangle(Trafo(10, 0, 0, 10, 0, 300))
        r21 = Rectangle(Trafo(10, 0, 0, 10, 100, 0))
        r22 = Rectangle(Trafo(10, 0, 0, 10, 100, 100))
        r23 = Rectangle(Trafo(10, 0, 0, 10, 100, 200))
        group = Group([r21, r22, r23])
        self.document.Insert(r1, self.layer)
        self.document.Insert(r2, self.layer)
        self.document.Insert(group, self.layer)
        self.document.Insert(r3, self.layer)
        self.document.Insert(r4, self.layer)

        self.editor.SelectObject([r2, r23, r4])

        # sanity check whether the selection is correct
        self.assertEquals(self.editor.SelectedObjects(),[r2, r23, r4] )

        # The actual operation to test
        self.editor.MoveSelectedToBottom()

        # The order of the children of the group should have changed
        # now.
        self.assertEquals(group.GetObjects(), [r23, r21, r22])
        self.assertEquals(self.layer.GetObjects(), [r2, r4, r1, group, r3])

        # The selection should still be the same, though.
        self.assertEquals(self.editor.SelectedObjects(), [r2, r23, r4])

        # check whether undo works properly
        self.document.Undo()

        # After the undo, the children of the layer and the group are
        # again in their original order and the selection of the editor
        # is still the same as before.
        self.assertEquals(group.GetObjects(), [r21, r22, r23])
        self.assertEquals(self.layer.GetObjects(), [r1, r2, group, r3, r4])
        self.assertEquals(self.editor.SelectedObjects(), [r2, r23, r4])



class EditorTransactionTest(DocEditorTestsWithMessages):

    """Run a test for a method of EditorWithSelection

    This class implements a generic test for a method of the
    EditorWithSelection class that is implemented as a transaction
    modifying the document.  The test tests the action and the undoing
    and redoing of the action.

    Derived classes have to implement the following methods which will
    be called automatically:

    action()

       Called without arguments by the test to perform the action to
       test.  Usually the implementation will be a simple call to the
       method of self.editor that is to be tested.

    check_before(after_undo)

       Called before the call to the action method and after the undo
       has been performed.  When called before the action the single
       boolean parameter is False and when called after the undo the
       parameter is True.  The implementation should test the state of
       the editor and document before the action resp. after the undo
       which is usually very similar, but there can be slight
       differences, especially when checking which messages were sent by
       the document, so the parameter can be used to distinguish between
       the two calls.

    check_after(after_redo)

       Called after the call to action and after the redo.  When called
       after the action the single boolean parameter is False and when
       called after the redo it is True.  The implementation should test
       the state of the editor and document after the action.  The
       boolean parameter servers the same purpose as in check_before.

    Additionally derived classes will usually want to extend the setUp
    method to perform additional initialization.

    The class arranges for two document messages to be subscribed:
    OBJECT_CHANGED and REDRAW.  Derived classes can check the messages
    with the methods check_redraw_messages and
    check_object_changed_messages.  The generic test implementation
    clears the list of received messages immediately before the action()
    call and before undo and redo so that the check_before and
    check_after methods only see the messages that were sent during the
    latest change to self.document.
    """


    subscribed_channels = [("document", [OBJECT_CHANGED, REDRAW])]

    def clear_messages(self):
        """Clear the list of received messages."""
        self.received_messages = []

    def messages_from_channel(self, channel):
        """Return a list with all messages of the given channel

        The return value is a list with all items of
        self.received_messages that were received on the given channel.
        The channel itself which is the last item in each of the tuples
        due to the way the subscription works in
        DocEditorTestsWithMessages is omitted, though.  The order of the
        items in the return value is the same as in
        self.received_messages.
        """
        return [message[:-1] for message in self.received_messages
                             if message[-1] == channel]

    def check_redraw_messages(self, rects):
        """Check whether there was one REDRAW message with the given rects.

        Fail the test unless:

         - There is exactly one REDRAW message

         - That message does not indicate that the entire window has to
           be redrawn.

         - The set of rectangles in the message is the same as that in
           the rects parameter of this method.  Both sets are sequences
           with arbitrary order so sorted copies of them are compared.
        """
        redraw_messages = self.messages_from_channel(REDRAW)
        self.assertEquals(len(redraw_messages), 1)

        # check that this is not a redraw_all message
        self.assertEquals(redraw_messages[0][0], 0)

        # check the rects.  sort them because the order doesn't matter.
        # Make sure to sort copies so that we avoid unexpected
        # modifications.
        msg_rects = list(redraw_messages[0][1])
        msg_rects.sort()
        rects = list(rects)
        rects.sort()
        self.assertEquals(msg_rects, rects)

    def check_object_changed_messages(self, expected):
        """Check the OBJECT_CHANGED messages that were received.

        Since the order of the messages is not defined sorted copies of
        self.messages_from_channel(OBJECT_CHANGED) and the given list of
        messages in the expected parameters are compared.  This also
        means that expected should be a sequence of expected messages in
        the format returned by self.messages_from_channel.
        """
        messages = self.messages_from_channel(OBJECT_CHANGED)
        messages.sort()

        expected = list(expected)
        expected.sort()
        self.assertEquals(messages, expected)

    def test(self):
        # Implement the generic test.  This not a doc-string because
        # under some circumstances the unittest framework uses it to
        # derive a title for the test which would be the same for all
        # tests implemented by the derived classes and that would be
        # more or less useless.
        self.check_before(False)

        try:
            self.clear_messages()
            self.action()
        except:
            self.fail("Exception during action(): %s"
                      % "".join(traceback.format_exception(*sys.exc_info())))

        self.check_after(False)

        self.clear_messages()
        self.document.Undo()
        self.check_before(True)

        self.clear_messages()
        self.document.Redo()
        self.check_after(True)

    def __call__(self, result):
        """Override to only execute real test implementations.

        A real test implementations as far as this method is concerned
        is an instance which has a check_after method.  The check is for
        check_after and not e.g. action because it's useful to implement
        the latter in a common baseclass for all tests for a given
        editor method and that base class would not be a real test
        either.  The check_after method on the other hand actually tests
        somethin and has to be much more specific to an actual test.
        """
        if hasattr(self, "check_after"):
            DocEditorTestsWithMessages.__call__(self, result)


class DuplicationTest(EditorTransactionTest):

    """Common base class for EditorWithSelection.DuplicateSelected() tests"""

    def action(self):
        self.editor.DuplicateSelected()


class TestDuplicateSimple(DuplicationTest):

    """Test EditorWithSelection.DuplicateSelected() very simple

    Test a very simple version of DuplicateSelected: two objects on one
    layer.  One is selected and then duplicated.  Test whether the new
    object is inserted immediately above the originally selected object.
    """

    def setUp(self):
        DuplicationTest.setUp(self)
        self.r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        self.r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(self.r1, self.layer)
        self.document.Insert(self.r2, self.layer)
        self.editor.SelectObject(self.r1)

    def check_before(self, after_undo):
        # check whether the structure of the document and the selection
        # before the action to test is as we expect it.  When called
        # before the action this serves a sanity check.  When called
        # after the undo this is the test that the undo reverted things
        # to their original state.
        self.assertEquals(self.editor.SelectedObjects(), [self.r1])
        self.assertEquals(self.layer.GetObjects(), [self.r1, self.r2])

        if after_undo:
            # One OBJECT_CHANGED message should have been sent by
            # the layer for the removed rectangle.
            self.check_object_changed_messages([(self.layer, CHILDREN,
                                               (REMOVED, (self.new_object,)))])
            # A REDRAW message should have been sent by the document with
            # the bounding rect of the new rectangle as parameter
            self.assertEquals(self.messages_from_channel(REDRAW),
                              [(0, [self.new_object.bounding_rect])])

    def action(self):
        self.editor.DuplicateSelected()

    def check_after(self, after_redo):
        # Now the duplicated object has been added to the layer just
        # above the originally selected object self.r1.
        objects = self.layer.GetObjects()
        self.new_object = objects[1]
        self.assertEquals(len(objects), 3)
        self.failUnless(objects[0] is self.r1)
        self.failUnless(objects[2] is self.r2)
        self.assertEquals(objects[1].Trafo(), Trafo(10, 0, 0, 10, 0, 0))

        # After duplication the new object should be selected
        self.assertEquals(self.editor.SelectedObjects(), [self.new_object])

        # One OBJECT_CHANGED message should have been sent by the
        # layer for the newly inserted rectangle.
        self.check_object_changed_messages([(self.layer, CHILDREN,
                                             (INSERTED, (self.new_object,)))])
        # A REDRAW message should have been sent by the document with
        # the bounding rect of the new rectangle as parameter
        self.assertEquals(self.messages_from_channel(REDRAW),
                          [(0, [self.new_object.bounding_rect,])])


class TestDuplicateTwoObjects(DuplicationTest):

    """Test EditorWithSelection.DuplicateSelected() two selected objects

    Two objects on one layer.  Both are selected and then duplicated.
    Test whether the each new object is inserted immediately above the
    original object.
    """

    def setUp(self):
        DuplicationTest.setUp(self)
        self.r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        self.r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(self.r1, self.layer)
        self.document.Insert(self.r2, self.layer)

        self.editor.SelectObject([self.r1, self.r2])

    def check_before(self, after_undo):
        # check whether the structure of the document and the selection
        # before the action to test is as we expect it.  When called
        # before the action this serves a sanity check.  When called
        # after the undo this is the test that the undo reverted things
        # to their original state.
        self.assertEquals(self.editor.SelectedObjects(), [self.r1, self.r2])
        self.assertEquals(self.layer.GetObjects(), [self.r1, self.r2])

        if after_undo:
            # A REDRAW message should have been sent by the document with
            # the bounding rects of the new rectangles as parameters.
            self.check_redraw_messages([self.r1_duplicate.bounding_rect,
                                        self.r2_duplicate.bounding_rect])

    def action(self):
        # The actual operation to test
        self.editor.DuplicateSelected()

    def check_after(self, after_redo):
        # After the action or after a redo, the duplicated objects have
        # been added to the layer layer just above their respective
        # originals so the order is self.r1, self.r1's duplicate,
        # self.r2, self.r2's duplicate
        objects = self.layer.GetObjects()
        self.r1_duplicate = objects[1]
        self.r2_duplicate = objects[3]
        self.assertEquals(len(objects), 4)
        self.failUnless(objects[0] is self.r1)
        self.failUnless(objects[2] is self.r2)
        self.assertEquals(self.r1_duplicate.Trafo(), Trafo(10, 0, 0, 10, 0, 0))
        self.assertEquals(self.r2_duplicate.Trafo(), Trafo(10, 0, 0, 10,0,100))

        # After duplication the new object should be selected
        self.assertEquals(self.editor.SelectedObjects(),
                          [self.r1_duplicate, self.r2_duplicate])

        # A REDRAW message should have been sent by the document with
        # the bounding rects of the new rectangles as parameters.
        self.check_redraw_messages([self.r1_duplicate.bounding_rect,
                                    self.r2_duplicate.bounding_rect])

        # FIXME: Add test for the OBJECT_CHANGED messages.  This is a
        # bit tricky since there's only one message listing one of the
        # new rectangles.  This is due to the
        # EditDocument.object_changed method which forwards only one
        # message per (object, what) pair.  That is probably a bug.


class TestDuplicateTwoObjectsTwoLayers(DuplicationTest):

    """Test EditorWithSelection.DuplicateSelected() two selected objects

    Three objects on two different layers.  Two objects on different
    layers are selected and then duplicated.  Test whether the each new
    object is inserted immediately above the original object in the same
    layer as the original.
    """

    def setUp(self):
        DuplicationTest.setUp(self)
        self.r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        self.r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.document.Insert(self.r1, self.layer)
        self.document.Insert(self.r2, self.layer)

        self.new_layer = self.document.AppendLayer("new layer")
        self.r3 = Rectangle(Trafo(100, 0, 0, 20, -100, 50))
        self.document.Insert(self.r3, self.new_layer)

        self.editor.SelectObject([self.r1, self.r3])

    def check_before(self, after_undo):
        self.assertEquals(self.editor.SelectedObjects(), [self.r1, self.r3])
        self.assertEquals(self.layer.GetObjects(), [self.r1, self.r2])
        self.assertEquals(self.new_layer.GetObjects(), [self.r3])

        if after_undo:
            # A REDRAW message should have been sent by the document with
            # the bounding rects of the new rectangles as parameters.
            self.check_redraw_messages([self.r1_duplicate.bounding_rect,
                                        self.r3_duplicate.bounding_rect])

            # Two OBJECT_CHANGED messages should have been sent by the
            # layers because the new objects have been removed by the
            # undo.
            self.check_object_changed_messages(
                [(self.layer, CHILDREN, (REMOVED, (self.r1_duplicate,))),
                 (self.new_layer, CHILDREN, (REMOVED, (self.r3_duplicate,)))])

    def action(self):
        # The actual operation to test
        self.editor.DuplicateSelected()

    def check_after(self, after_redo):
        # The duplicate of self.r1 should be in self.layer immediately
        # above self.r1.  The duplicate of self.r3 is in self.new_layer
        # immediately above self.r3.
        objects = self.layer.GetObjects()
        self.r1_duplicate = objects[1]
        self.assertEquals(len(objects), 3)
        self.failUnless(objects[0] is self.r1)
        self.failUnless(objects[2] is self.r2)
        self.assertEquals(self.r1_duplicate.Trafo(), Trafo(10, 0, 0, 10, 0, 0))

        objects = self.new_layer.GetObjects()
        self.r3_duplicate = objects[1]
        self.assertEquals(len(objects), 2)
        self.failUnless(objects[0] is self.r3)
        self.assertEquals(self.r3_duplicate.Trafo(),
                          Trafo(100, 0, 0, 20, -100, 50))

        # After duplication the new object should be selected
        self.assertEquals(self.editor.SelectedObjects(),
                          [self.r1_duplicate, self.r3_duplicate])

        # A REDRAW message should have been sent by the document with
        # the bounding rects of the new rectangles as parameters.
        self.check_redraw_messages([self.r1_duplicate.bounding_rect,
                                    self.r3_duplicate.bounding_rect])

        # Two OBJECT_CHANGED messages should have been sent by the
        # layers for the newly inserted rectangles.
        self.check_object_changed_messages(
            [(self.layer, CHILDREN, (INSERTED, (self.r1_duplicate,))),
             (self.new_layer, CHILDREN, (INSERTED, (self.r3_duplicate,)))])


class TestDuplicateInGroup(DuplicationTest):

    """Test EditorWithSelection.DuplicateSelected() with grouped objects

    Two objects in a group.  One is selected and then duplicated.  Test
    whether the new object is inserted immediately above the originally
    selected object but in the same group.
    """

    def setUp(self):
        DuplicationTest.setUp(self)
        self.r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        self.r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.group = Group([self.r1, self.r2])
        self.document.Insert(self.group, self.layer)

        self.editor.SelectObject(self.r1)

    def check_before(self, after_undo):
        self.assertEquals(self.editor.SelectedObjects(), [self.r1])
        self.assertEquals(self.group.GetObjects(), [self.r1, self.r2])
        self.assertEquals(self.editor.SelectedObjects(), [self.r1])

        if after_undo:
            # A REDRAW message should have been sent by the document with
            # the bounding rect of the new rectangle as parameter.
            self.check_redraw_messages([self.new_object.bounding_rect])

            # One OBJECT_CHANGED messages should have been sent by the
            # group because the new object has been removed again by the
            # undo
            self.check_object_changed_messages(
                [(self.group, CHILDREN, (REMOVED, (self.new_object,)))])

    def action(self):
        # The actual operation to test
        self.editor.DuplicateSelected()

    def check_after(self, after_redo):
        # Now the duplicated object has been added to the layer just
        # above the originally selected object self.r1.
        objects = self.group.GetObjects()
        self.new_object = objects[1]
        self.assertEquals(len(objects), 3)
        self.failUnless(objects[0] is self.r1)
        self.failUnless(objects[2] is self.r2)
        self.assertEquals(self.new_object.Trafo(), Trafo(10, 0, 0, 10, 0, 0))

        # After duplication the new object should be selected
        self.assertEquals(self.editor.SelectedObjects(), [self.new_object])

        # A REDRAW message should have been sent by the document with
        # the bounding rect of the new rectangle as parameter.
        self.check_redraw_messages([self.new_object.bounding_rect])

        # One OBJECT_CHANGED messages should have been sent by the group
        # for the newly inserted rectangle.
        self.check_object_changed_messages(
            [(self.group, CHILDREN, (INSERTED, (self.new_object,)))])



class RemovalTest(EditorTransactionTest):

    """Common base class for EditorWithSelection.RemoveSelected() tests"""

    def action(self):
        self.editor.RemoveSelected()


class TestRemoveSimple(RemovalTest):
    """Test EditorWithSelection.RemoveSelected() simple

    Create three rectangles in the current layer.  Select one of them.
    RemoveSelected should remove the selected object.  The others should
    still be there.  The selection should be empty afterwards because
    the selected object has been removed.
    """

    def setUp(self):
        RemovalTest.setUp(self)
        self.r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        self.r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.document.Insert(self.r1, self.layer)
        self.document.Insert(self.r2, self.layer)
        self.document.Insert(self.r3, self.layer)

        self.editor.SelectObject(self.r2)

    def check_before(self, after_undo):
        self.assertEquals(self.layer.GetObjects(), [self.r1, self.r2, self.r3])
        self.assertEquals(self.editor.SelectedObjects(), [self.r2])

        if after_undo:
            # The area defined by the bounding box of the selected
            # object has to be redrawn after the undo
            self.check_redraw_messages([self.r2.bounding_rect])

            # One OBJECT_CHANGED message for the removed rectangle which
            # was reinserted by the undo
            self.check_object_changed_messages([(self.layer, CHILDREN,
                                                 (INSERTED, (self.r2,)))])

    def check_after(self, after_redo):
        # self.r2 should be gone now
        self.assertEquals(self.layer.GetObjects(), [self.r1, self.r3])

        # The selection should be empty now because the selected objects
        # have been removed
        self.assertEquals(self.editor.SelectedObjects(), [])

        # The area defined by the bounding box of the selected object
        # has to be redrawn
        self.check_redraw_messages([self.r2.bounding_rect])

        # One OBJECT_CHANGED message for the removed rectangle.
        self.check_object_changed_messages([(self.layer, CHILDREN,
                                             (REMOVED, (self.r2,)))])

class TestRemoveFromGroup(RemovalTest):

    """Test EditorWithSelection.RemoveSelected() from group

    Create a group with three rectangles and select one of them.
    RemoveSelected should remove the selected object only.  The group
    and the other rectangles should still be there.  The selection
    should be empty afterwards because the selected object has been
    removed.
    """

    def setUp(self):
        RemovalTest.setUp(self)
        self.r1 = Rectangle(Trafo(10, 0, 0, 10, 0, 0))
        self.r2 = Rectangle(Trafo(10, 0, 0, 10, 0, 100))
        self.r3 = Rectangle(Trafo(10, 0, 0, 10, 0, 200))
        self.group = Group([self.r1, self.r2, self.r3])
        self.document.Insert(self.group, self.layer)
        self.editor.SelectObject(self.r2)

    def check_before(self, after_undo):
        self.assertEquals(self.editor.SelectedObjects(), [self.r2])
        self.assertEquals(self.group.GetObjects(), [self.r1, self.r2, self.r3])
        self.assertEquals(self.editor.SelectedObjects(), [self.r2])

        if after_undo:
            # The area defined by the bounding box of the selected
            # object has to be redrawn after the undo
            self.check_redraw_messages([self.r2.bounding_rect])

            # One OBJECT_CHANGED message for the removed rectangle which
            # was reinserted by the undo
            self.check_object_changed_messages([(self.group, CHILDREN,
                                                 (INSERTED, (self.r2,)))])

    def check_after(self, after_redo):
        # self.r2 should be gone now
        self.assertEquals(self.group.GetObjects(), [self.r1, self.r3])
        # but self.group should still be there
        self.assertEquals(self.layer.GetObjects(), [self.group])

        # The selection should be empty now because the selected objects
        # have been removed
        self.assertEquals(self.editor.SelectedObjects(), [])

        # The area defined by the bounding box of the selected object
        # has to be redrawn
        self.check_redraw_messages([self.r2.bounding_rect])

        # One OBJECT_CHANGED message for the removed rectangle.
        self.check_object_changed_messages([(self.group, CHILDREN,
                                             (REMOVED, (self.r2,)))])


if __name__ == "__main__":
    unittest.main()
