# Sketch - A Python-based interactive drawing program
# Copyright (C) 2003 by Bernhard Herzog
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
Test cases for the selection
"""

__version__ = "$Revision: 1.1 $"
# $Source: /cvs/gnome/sketch/test/test_selection.py,v $
# $Id: test_selection.py,v 1.1 2003/01/12 18:43:30 bherzog Exp $

import unittest

import support
support.add_sketch_dir_to_path()

from Sketch.Editor.selection import Selection
from Sketch.Editor import InteractiveDocument
from Sketch.Graphics import Rectangle
from Sketch import Trafo


class Rectangle1(Rectangle):

    """Rectangle with a simple Info text"""

    def Info(self):
        """Override the Info method to return a simple string"""
        return "Rectangle"


class Rectangle2(Rectangle):

    """Rectangle with a more complex Info text"""

    def Info(self):
        """Override the Info method to return a string and a dict."""
        return "%(name)s", {"name": "Rectangle"}



class TestSelection(unittest.TestCase):

    def setUp(self):
        """Create a document with three rectangles and two layers.

        The layers are 'Background' and 'Foreground' and are bound to
        layer1 and layer2 repectively.

        The rectangles are rect11 and rect12 on layer1 and self.rect2 on
        layer2. The rectangles are actually instances of Rectangle1 resp
        rectangle2 so that it's easier to compare InfoText results.

        The document is bound to doc.
        """
        # First build a document
        self.doc = InteractiveDocument(create_layer = 1)

        # Create two layers
        self.layer1 = self.doc.AppendLayer("Background")
        self.layer2 = self.doc.AppendLayer("Foreground")

        # Insert a rectangle into each
        self.rect11 = Rectangle1()
        self.layer1.Insert(self.rect11)
        self.rect12 = Rectangle1()
        self.layer1.Insert(self.rect12)

        self.rect2 = Rectangle2()
        self.layer2.Insert(self.rect2)

    def test_info_text_no_object(self):
        """Test Selection.InfoText with no object"""
        # Create the selection. Initially selections are empty
        # all
        sel = Selection(self.doc)

        self.assertEquals(sel.InfoText(), "No Selection")

    def test_info_text_one_object_simple_info(self):
        """Test Selection.InfoText with one object with simple Info"""
        # Create the selection
        sel = Selection(self.doc)

        # Now select a rect
        sel.SetSelection([self.rect11])

        text, values = sel.InfoText()
        self.assertEquals(text % values, "Rectangle on `Background'")

    def test_info_text_one_object_complex_info(self):
        """Test Selection.InfoText with one object with complex Info"""
        # Create the selection
        sel = Selection(self.doc)

        # Now select a rect
        sel.SetSelection([self.rect2])

        text, values = sel.InfoText()
        self.assertEquals(text % values, "Rectangle on `Foreground'")

    def test_info_text_several_layers(self):
        """Test Selection.InfoText with two objects on separate layers"""
        # Create the selection
        sel = Selection(self.doc)

        # Now select two rects
        sel.SetSelection([self.rect11, self.rect2])

        self.assertEquals(sel.InfoText(), "2 objects on several layers")

    def test_info_text_several_objects_one_layer(self):
        """Test Selection.InfoText with two objects on the same layer"""
        # Create the selection
        sel = Selection(self.doc)

        # Now select two rects
        sel.SetSelection([self.rect11, self.rect12])

        self.assertEquals(sel.InfoText(), "2 objects on `Background'")
